@@ -10,40 +10,24 @@
  * WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
  */
 
-#![cfg_attr(feature = "aggressive_lint", deny(warnings))]
-
-use std::io::prelude::*;
 use std::path::Path;
 
-use super::rusoto::*;
-
-use super::ObjectInfo;
-use super::Result;
-use super::S3PathParam;
+use crate::ops::sync::GlobFilter;
+use crate::rusoto::*;
+use crate::HeadObjectInfo;
+use crate::Result;
+use crate::S3PathParam;
 
 #[tokio::main]
 pub async fn head_object<T>(
     s3: &T,
     bucket: impl AsRef<str>,
     key: impl AsRef<str>,
-) -> Result<Option<ObjectInfo>>
+) -> Result<Option<HeadObjectInfo>>
 where
     T: S3 + Send,
 {
-    super::head_object(s3, bucket, key).await
-}
-
-#[tokio::main]
-pub async fn abort_upload<T>(
-    s3: &T,
-    bucket: impl AsRef<str>,
-    key: impl AsRef<str>,
-    upload_id: impl AsRef<str>,
-) -> Result<()>
-where
-    T: S3 + Send,
-{
-    super::abort_upload(s3, bucket, key, upload_id).await
+    crate::head_object(s3, bucket, key).await
 }
 
 #[tokio::main]
@@ -56,25 +40,9 @@ pub async fn upload<T>(
 where
     T: S3 + Send + Clone,
 {
-    super::upload(s3, bucket, key, file).await
+    crate::upload(s3, bucket, key, file).await
 }
 
-#[tokio::main]
-pub async fn upload_from_reader<T, R>(
-    s3: &T,
-    bucket: impl AsRef<str>,
-    key: impl AsRef<str>,
-    reader: R,
-    file_size: u64,
-) -> Result<()>
-where
-    T: S3 + Send + Clone,
-    R: Read + Send + 'static,
-{
-    super::upload_from_reader(s3, bucket, key, reader, file_size, None).await
-}
-
-#[cfg(feature = "compression")]
 #[tokio::main]
 pub async fn upload_compressed<T>(
     s3: &T,
@@ -85,7 +53,7 @@ pub async fn upload_compressed<T>(
 where
     T: S3 + Send + Clone,
 {
-    super::upload_compressed(s3, bucket, key, file).await
+    crate::upload_compressed(s3, bucket, key, file).await
 }
 
 #[tokio::main]
@@ -98,12 +66,11 @@ pub async fn download<T>(
 where
     T: S3 + Sync + Send + Clone,
 {
-    super::download(s3, bucket, key, file).await
+    crate::download(s3, bucket, key, file).await
 }
 
-#[cfg(feature = "compression")]
 #[tokio::main]
-pub async fn download_decompressed<T>(
+pub async fn download_with_transparent_decompression<T>(
     s3: &T,
     bucket: impl AsRef<str>,
     key: impl AsRef<str>,
@@ -112,7 +79,7 @@ pub async fn download_decompressed<T>(
 where
     T: S3 + Sync + Send + Clone,
 {
-    super::download_decompressed(s3, bucket, key, file).await
+    crate::download_with_transparent_decompression(s3, bucket, key, file).await
 }
 
 #[tokio::main]
@@ -120,23 +87,13 @@ pub async fn sync<T>(
     s3: &T,
     source: S3PathParam,
     destination: S3PathParam,
-    includes: Option<&[impl AsRef<str>]>,
-    excludes: Option<&[impl AsRef<str>]>,
-    #[cfg(feature = "compression")] compressed: bool,
+    filters: Option<&[GlobFilter]>,
+    transparent_compression: bool,
 ) -> Result<()>
 where
     T: S3 + Sync + Send + Clone,
 {
-    super::sync(
-        s3,
-        source,
-        destination,
-        includes,
-        excludes,
-        #[cfg(feature = "compression")]
-        compressed,
-    )
-    .await
+    crate::sync(s3, source, destination, filters, transparent_compression).await
 }
 
 #[tokio::main]
@@ -148,10 +105,10 @@ pub async fn list_objects<T>(
 where
     T: S3 + Send,
 {
-    super::list_objects(s3, bucket, key).await
+    crate::list_objects(s3, bucket, key).await
 }
 
 #[tokio::main]
 pub async fn compute_etag(path: impl AsRef<Path>) -> Result<String> {
-    super::compute_etag(path).await
+    crate::compute_etag(path).await
 }
