@@ -10,75 +10,81 @@
  * WARRANTIES OF MERCHANTABILITY AND/OR FITNESS FOR A PARTICULAR PURPOSE.
  */
 
-#![cfg_attr(feature = "aggressive_lint", deny(warnings))]
-
-use std::path::{Path, PathBuf};
+use std::{
+    borrow::Cow,
+    path::{Path, PathBuf},
+};
 
 use futures::{Future, Stream, StreamExt, TryStreamExt};
 use glob::Pattern;
 use log::{debug, info, warn};
 use log_derive::logfn;
+use tokio::{
+    fs::{self, File},
+    io::BufReader,
+};
 use walkdir::WalkDir;
 
 use crate::{
+    compression::compress_to_tempfile,
     compute_etag,
     config::Config,
     errors::{Error, Result},
-    handle_dispatch_error, head_object_request, list_objects_stream,
+    handle_dispatch_error, list_objects_stream,
     rusoto::*,
+    tempfile::TEMP_FILE_PREFIX,
     types::ListingMetadata,
     types::{S3ListingItem, S3PathParam},
 };
 
-type MapEtagResult = Result<(PathBuf, Result<String>, Option<ListingMetadata>)>;
+struct MappedPathResult {
+    file_path: Box<dyn AsRef<Path>>,
+    source_path: PathBuf,
+    local_etag: Result<String>,
+    metadata: Option<ListingMetadata>,
+}
 
+type MapPathResult = Result<MappedPathResult>;
+
+#[derive(Debug, Clone)]
+pub enum GlobFilter {
+    Include(Pattern),
+    Exclude(Pattern),
+}
+
+/// Syncs between S3 prefixes and local directories
+///
+/// # Arguments
+///
+/// * `s3` - S3 client
+/// * `source` - S3 prefix or local directory to sync from
+/// * `destination` - S3 prefix or local directory to sync to
+/// * `glob_filter` - An (optional) slice of filters that specify whether files
+///                   should be included or not. These are processed in order,
+///                   with the first matching filter determining whether the
+///                   file will be included or excluded. If not supplied, then
+///                   all files will be synced.
 #[logfn(err = "ERROR")]
 pub async fn sync<T>(
     s3: &T,
     source: S3PathParam,
     destination: S3PathParam,
-    includes: Option<&[impl AsRef<str>]>,
-    excludes: Option<&[impl AsRef<str>]>,
-    #[cfg(feature = "compression")] compressed: bool,
+    glob_filters: Option<&[GlobFilter]>,
+    compressed: bool,
 ) -> Result<()>
 where
     T: S3 + Sync + Send + Clone,
 {
-    #[cfg(not(feature = "compression"))]
-    let compressed = false;
-
-    let mut glob_excludes: Vec<Pattern> = vec![];
-    let mut glob_includes: Vec<Pattern> = vec![];
-
-    if let Some(excludes) = excludes {
-        for exclude in excludes {
-            let exclude = exclude.as_ref();
-            match Pattern::new(exclude) {
-                Err(e) => {
-                    return Err(Error::GlobPatternError(e));
-                }
-                Ok(p) => {
-                    glob_excludes.push(p);
-                }
+    let filters: Vec<GlobFilter> = match glob_filters {
+        Some(filters) => {
+            let mut filters = filters.to_vec();
+            if !filters.iter().any(|x| matches!(x, GlobFilter::Include(_))) {
+                filters.push(GlobFilter::Include(Pattern::new("*")?));
             }
+            filters
         }
-    }
-
-    if let Some(includes) = includes {
-        for include in includes {
-            let include = include.as_ref();
-            match Pattern::new(include) {
-                Err(e) => {
-                    return Err(Error::GlobPatternError(e));
-                }
-                Ok(p) => {
-                    glob_includes.push(p);
-                }
-            }
-        }
-    } else {
-        glob_includes.push(Pattern::new("*")?);
-    }
+        None => vec![GlobFilter::Include(Pattern::new("*")?)],
+    };
 
     match (source, destination) {
         (S3PathParam::Local { path }, S3PathParam::Bucket { bucket, key }) => {
@@ -89,16 +95,7 @@ where
                 key
             );
 
-            sync_local_to_remote(
-                s3,
-                &bucket,
-                &key,
-                &path,
-                &glob_includes,
-                &glob_excludes,
-                compressed,
-            )
-            .await?;
+            sync_local_to_remote(s3, &bucket, &key, &path, &filters, compressed).await?;
         }
         (S3PathParam::Bucket { bucket, key }, S3PathParam::Local { path }) => {
             info!(
@@ -108,16 +105,7 @@ where
                 key
             );
 
-            sync_remote_to_local(
-                s3,
-                &bucket,
-                &key,
-                &path,
-                &glob_includes,
-                &glob_excludes,
-                compressed,
-            )
-            .await?;
+            sync_remote_to_local(s3, &bucket, &key, &path, &filters, compressed).await?;
         }
         (
             S3PathParam::Bucket {
@@ -140,8 +128,7 @@ where
                 &source_key,
                 &destination_bucket,
                 &destination_key,
-                &glob_includes,
-                &glob_excludes,
+                &filters,
             )
             .await?;
         }
@@ -153,25 +140,25 @@ where
     Ok(())
 }
 
-fn process_globs<'a, P: AsRef<Path> + 'a>(
-    path: P,
-    glob_includes: &[Pattern],
-    glob_excludes: &[Pattern],
-) -> Option<P> {
+fn process_globs<'a, P: AsRef<Path> + 'a>(path: P, filters: &[GlobFilter]) -> Option<P> {
     let mut excluded = false;
     let mut included = false;
     {
         let path = path.as_ref();
-        for pattern in glob_excludes {
-            if pattern.matches(path.to_string_lossy().as_ref()) {
-                excluded = true;
-                break;
-            }
-        }
-        for pattern in glob_includes {
-            if pattern.matches(path.to_string_lossy().as_ref()) {
-                included = true;
-                break;
+        for pattern in filters {
+            match pattern {
+                GlobFilter::Include(filter) => {
+                    if filter.matches(path.to_string_lossy().as_ref()) {
+                        included = true;
+                        break;
+                    }
+                }
+                GlobFilter::Exclude(filter) => {
+                    if filter.matches(path.to_string_lossy().as_ref()) {
+                        excluded = true;
+                        break;
+                    }
+                }
             }
         }
     }
@@ -184,8 +171,7 @@ fn process_globs<'a, P: AsRef<Path> + 'a>(
 
 fn create_dirent_stream<'a>(
     directory: &'a Path,
-    glob_includes: &'a [Pattern],
-    glob_excludes: &'a [Pattern],
+    filters: &'a [GlobFilter],
 ) -> impl Stream<Item = Result<(String, Option<ListingMetadata>)>> + 'a {
     async_stream::stream! {
         for entry in WalkDir::new(directory) {
@@ -195,6 +181,9 @@ fn create_dirent_stream<'a>(
                 yield Err(entry.err().unwrap().into());
                 return;
             };
+            if entry.file_name().to_string_lossy().contains(TEMP_FILE_PREFIX) {
+                continue;
+            }
             let metadata = entry.metadata();
             let stat = if let Ok(stat) = metadata {
                 stat
@@ -211,7 +200,7 @@ fn create_dirent_stream<'a>(
             }
             let path = entry.path();
             debug!("local path={}", path.display());
-            if process_globs(&path, glob_includes, glob_excludes).is_some() {
+            if process_globs(&path, filters).is_some() {
                 yield Ok((path.to_string_lossy().into(), ListingMetadata::none()));
             }
         }
@@ -225,74 +214,65 @@ enum SyncCompressionDirection {
     Down,
 }
 
-fn map_paths_to_etags<StreamT>(
+fn translate_paths<StreamT>(
     input_stream: StreamT,
     compressed: SyncCompressionDirection,
-) -> impl Stream<Item = impl Future<Output = MapEtagResult>>
+) -> impl Stream<Item = impl Future<Output = MapPathResult>>
 where
     StreamT: Stream<Item = Result<(String, Option<ListingMetadata>)>>,
 {
     use std::str::FromStr;
     input_stream.map(move |params| async move {
         let (path, metadata) = params?;
-        let path = PathBuf::from_str(&path)?;
-        #[cfg(feature = "compression")]
-        {
-            use crate::compression::{
-                compress_and_replace, compress_to_tempfile, compressed_path_to_path,
-            };
+        let source_path = PathBuf::from_str(&path)?;
+        let file_path: Box<dyn AsRef<Path>> = Box::new(path);
 
-            let (path, local_etag) = match compressed {
-                SyncCompressionDirection::None => {
-                    let local_etag = compute_etag(&path).await;
-                    (path, local_etag)
-                }
-                SyncCompressionDirection::Up => {
-                    if path.extension().map(|e| e == "gz").unwrap_or(false) {
-                        let local_etag = compute_etag(&path).await;
-                        (path, local_etag)
-                    } else {
-                        info!("compressing and replacing: {}", path.display());
-                        let path = compress_and_replace(path).await?;
-                        let local_etag = compute_etag(&path).await;
-                        (path, local_etag)
-                    }
-                }
-                SyncCompressionDirection::Down => {
-                    if path.extension().map(|e| e == "gz").unwrap_or(false) {
-                        let uncompressed_path =
-                            compressed_path_to_path(&path).expect("Path should be compressed");
-
-                        if !uncompressed_path.exists() {
-                            (path, Err(Error::ETagNotPresent))
-                        } else {
-                            // Check if we already have a copy locally
-                            // of the uncompressed file by recompressing
-                            // the local file to see if it matches with
-                            // the compressed version
-                            let (temp_compressed, _) =
-                                compress_to_tempfile(&uncompressed_path).await?;
-                            let local_etag = compute_etag(&temp_compressed).await;
-                            (path, local_etag)
-                        }
-                    } else {
-                        let local_etag = compute_etag(&path).await;
-                        (path, local_etag)
-                    }
-                }
-            };
-
-            Ok((path, local_etag, metadata))
-        }
-        #[cfg(not(feature = "compression"))]
-        {
-            if matches!(compressed, SyncCompressionDirection::None) {
-                let local_etag = compute_etag(&path).await;
-                Ok((path, local_etag, metadata))
-            } else {
-                panic!("compression feature not enabled");
+        let (file_path, source_path, local_etag) = match compressed {
+            SyncCompressionDirection::None => {
+                let local_etag = compute_etag(&source_path).await;
+                (file_path, source_path, local_etag)
             }
-        }
+            SyncCompressionDirection::Up => {
+                // If we're syncing up with compression, then everything
+                // should be compressed
+                let (tmp, _) = compress_to_tempfile(&source_path).await?;
+                let local_etag = compute_etag(tmp.path()).await;
+                (tmp.into_path(), source_path, local_etag)
+            }
+            SyncCompressionDirection::Down => {
+                // if we're syncing down with compression, then there
+                // could be both esthri compressed files and non
+                // compressed files within the prefix. We should only
+                // try to decompress the esthri compressed files
+                if metadata
+                    .as_ref()
+                    .expect("Should have metadata")
+                    .esthri_compressed
+                {
+                    if !source_path.exists() {
+                        (file_path, source_path, Err(Error::ETagNotPresent))
+                    } else {
+                        // Check if we already have a copy locally
+                        // of the uncompressed file by recompressing
+                        // the local file to see if it matches with
+                        // the compressed version
+                        let (tmp, _) = compress_to_tempfile(&source_path).await?;
+                        let local_etag = compute_etag(tmp.path()).await;
+                        (file_path, source_path, local_etag)
+                    }
+                } else {
+                    let local_etag = compute_etag(&source_path).await;
+                    (file_path, source_path, local_etag)
+                }
+            }
+        };
+
+        Ok(MappedPathResult {
+            file_path,
+            source_path,
+            local_etag,
+            metadata,
+        })
     })
 }
 
@@ -302,12 +282,13 @@ fn local_to_remote_sync_tasks<ClientT, StreamT>(
     key: String,
     directory: PathBuf,
     dirent_stream: StreamT,
+    transparent_compression: bool,
 ) -> impl Stream<Item = impl Future<Output = Result<()>>>
 where
     ClientT: S3 + Send + Clone,
-    StreamT: Stream<Item = MapEtagResult>,
+    StreamT: Stream<Item = MapPathResult>,
 {
-    use super::upload::upload;
+    use super::upload::upload_from_reader;
     dirent_stream
         .map(move |entry| {
             (
@@ -320,13 +301,21 @@ where
         })
         .map(move |clones| async move {
             let (s3, bucket, key, directory, entry) = clones;
-            let (path, local_etag, _metadata) = entry;
-            let path = Path::new(&path);
+            let MappedPathResult {
+                file_path: filepath,
+                source_path,
+                local_etag,
+                ..
+            } = entry;
+            let path = Path::new(&source_path);
             let remote_path = Path::new(&key);
             let stripped_path = path.strip_prefix(&directory);
             let stripped_path = match stripped_path {
                 Err(e) => {
-                    warn!("unexpected: failed to strip prefix: {}", e);
+                    warn!(
+                        "unexpected: failed to strip prefix: {}, {:?}, {:?}",
+                        e, &path, &directory
+                    );
                     return Ok(());
                 }
                 Ok(result) => result,
@@ -335,7 +324,13 @@ where
             let remote_path = remote_path.to_string_lossy();
             debug!("checking remote: {}", remote_path);
             let local_etag = local_etag?;
-            let object_info = head_object_request(&s3, &bucket, &remote_path).await?;
+            let object_info = head_object_request(&s3, &bucket, &remote_path, None).await?;
+            let metadata = if transparent_compression {
+                Some(crate::compression::compressed_file_metadata())
+            } else {
+                None
+            };
+
             if let Some(object_info) = object_info {
                 let remote_etag = object_info.e_tag;
                 if remote_etag != local_etag {
@@ -343,7 +338,10 @@ where
                         "etag mis-match: {}, remote_etag={}, local_etag={}",
                         remote_path, remote_etag, local_etag
                     );
-                    upload(&s3, bucket, &remote_path, &path).await?;
+                    let f = File::open(&*filepath).await?;
+                    let reader = BufReader::new(f);
+                    let size = fs::metadata(&*filepath).await?.len();
+                    upload_from_reader(&s3, bucket, remote_path, reader, size, metadata).await?;
                 } else {
                     debug!(
                         "etags matched: {}, remote_etag={}, local_etag={}",
@@ -352,7 +350,10 @@ where
                 }
             } else {
                 info!("file did not exist remotely: {}", remote_path);
-                upload(&s3, bucket, &remote_path, &path).await?;
+                let f = File::open(&*filepath).await?;
+                let reader = BufReader::new(f);
+                let size = fs::metadata(&*filepath).await?.len();
+                upload_from_reader(&s3, bucket, remote_path, reader, size, metadata).await?;
             }
             Ok(())
         })
@@ -363,8 +364,7 @@ async fn sync_local_to_remote<T>(
     bucket: &str,
     key: &str,
     directory: impl AsRef<Path>,
-    glob_includes: &[Pattern],
-    glob_excludes: &[Pattern],
+    filters: &[GlobFilter],
     compressed: bool,
 ) -> Result<()>
 where
@@ -372,20 +372,21 @@ where
 {
     let directory = directory.as_ref();
     let task_count = Config::global().concurrent_sync_tasks();
-    let dirent_stream = create_dirent_stream(directory, glob_includes, glob_excludes);
+    let dirent_stream = create_dirent_stream(directory, filters);
     let compression = if compressed {
         SyncCompressionDirection::Up
     } else {
         SyncCompressionDirection::None
     };
 
-    let etag_stream = map_paths_to_etags(dirent_stream, compression).buffer_unordered(task_count);
+    let etag_stream = translate_paths(dirent_stream, compression).buffer_unordered(task_count);
     let sync_tasks = local_to_remote_sync_tasks(
         s3.clone(),
         bucket.into(),
         key.into(),
         directory.into(),
         etag_stream,
+        compressed,
     );
 
     sync_tasks
@@ -411,7 +412,7 @@ where
     let res = handle_dispatch_error(|| async {
         let cor = CopyObjectRequest {
             bucket: dest_bucket.to_string(),
-            copy_source: format!("{}/{}", source_bucket.to_string(), &file_name),
+            copy_source: format!("{}/{}", source_bucket, file_name),
             key: file_name.replace(source_key, dest_key),
             ..Default::default()
         };
@@ -429,8 +430,7 @@ async fn sync_across<T>(
     source_prefix: &str,
     dest_bucket: &str,
     destination_key: &str,
-    glob_includes: &[Pattern],
-    glob_excludes: &[Pattern],
+    filters: &[GlobFilter],
 ) -> Result<()>
 where
     T: S3 + Send,
@@ -440,12 +440,13 @@ where
     while let Some(from_entries) = stream.try_next().await? {
         for entry in from_entries {
             if let S3ListingItem::S3Object(src_object) = entry {
-                let path = process_globs(&src_object.key, glob_includes, glob_excludes);
+                let path = process_globs(&src_object.key, filters);
 
                 if let Some(_accept) = path {
                     let mut should_copy_file: bool = true;
                     let new_file = src_object.key.replace(source_prefix, destination_key);
-                    let dest_object_info = head_object_request(s3, dest_bucket, &new_file).await?;
+                    let dest_object_info =
+                        head_object_request(s3, dest_bucket, &new_file, None).await?;
 
                     if let Some(dest_object) = dest_object_info {
                         if dest_object.e_tag == src_object.e_tag {
@@ -477,39 +478,67 @@ fn flattened_object_listing<'a, ClientT>(
     bucket: &'a str,
     key: &'a str,
     directory: &'a Path,
-    glob_includes: &'a [Pattern],
-    glob_excludes: &'a [Pattern],
+    filters: &'a [GlobFilter],
+    transparent_decompress: bool,
 ) -> impl Stream<Item = Result<(String, Option<ListingMetadata>)>> + 'a
 where
     ClientT: S3 + Send + Clone,
 {
+    let prefix = if key.ends_with('/') {
+        Cow::Borrowed(key)
+    } else {
+        Cow::Owned(format!("{}/", key))
+    };
+
     async_stream::stream! {
-        let mut stream = list_objects_stream(s3, bucket, key);
+        let mut stream = list_objects_stream(s3, bucket, prefix);
         loop {
-            let entries_result = stream.try_next().await;
-            if let Ok(entries_option) = entries_result {
-                if let Some(entries) = entries_option {
-                    for entry in entries {
-                        let entry = entry.unwrap_object();
-                        debug!("key={}", entry.key);
-                        let path_result = Path::new(&entry.key).strip_prefix(key);
-                        if let Ok(s3_suffix) = path_result {
-                            if process_globs(&s3_suffix, glob_includes, glob_excludes).is_some() {
-                                let local_path: String = directory.join(&s3_suffix).to_string_lossy().into();
-                                let s3_suffix = s3_suffix.to_string_lossy().into();
-                                yield Ok((local_path, ListingMetadata::some(s3_suffix, entry.e_tag)));
-                            }
-                        } else {
-                            yield Err(path_result.err().unwrap().into());
-                            return;
-                        }
+            let entries = match stream.try_next().await {
+                Ok(Some(entries)) => entries,
+                Ok(None) => break,
+                Err(err) => {
+                    yield Err(err);
+                    return;
+                }
+            };
+            let mut entries = futures::stream::iter(entries)
+                .map(|entry| async move {
+                    let entry = entry.unwrap_object();
+                    debug!("key={}", entry.key);
+                    let compressed = if transparent_decompress {
+                        head_object_request(s3, bucket, &entry.key, None)
+                            .await?
+                            .expect("No head info?")
+                            .is_esthri_compressed()
+                    } else {
+                        false
+                    };
+                    Ok((entry, compressed))
+                })
+                .buffered(Config::global().concurrent_sync_tasks());
+            loop {
+                let (entry, compressed) = match entries.try_next().await {
+                    Ok(Some(entry)) => entry,
+                    Ok(None) => break,
+                    Err(err) => {
+                        yield Err(err);
+                        return;
+                    }
+                };
+                let path_result = Path::new(&entry.key).strip_prefix(key);
+                if let Ok(s3_suffix) = path_result {
+                    if process_globs(&s3_suffix, filters).is_some() {
+                        let local_path: String = directory.join(&s3_suffix).to_string_lossy().into();
+                        let s3_suffix = s3_suffix.to_string_lossy().into();
+                        yield Ok((
+                            local_path,
+                            ListingMetadata::some(s3_suffix, entry.e_tag, compressed),
+                        ));
                     }
                 } else {
-                    break;
+                    yield Err(path_result.err().unwrap().into());
+                    return;
                 }
-            } else {
-                yield Err(entries_result.err().unwrap());
-                return;
             }
         }
     }
@@ -521,13 +550,12 @@ fn remote_to_local_sync_tasks<ClientT, StreamT>(
     key: String,
     directory: PathBuf,
     input_stream: StreamT,
-    decompress: bool,
+    transparent_decompress: bool,
 ) -> impl Stream<Item = impl Future<Output = Result<()>>>
 where
     ClientT: S3 + Sync + Send + Clone,
-    StreamT: Stream<Item = MapEtagResult>,
+    StreamT: Stream<Item = MapPathResult>,
 {
-    use super::download::download_with_dir;
     input_stream
         .map(move |entry| {
             (
@@ -535,20 +563,25 @@ where
                 bucket.clone(),
                 key.clone(),
                 directory.clone(),
-                decompress,
+                transparent_decompress,
                 entry.unwrap(),
             )
         })
         .map(
             |(s3, bucket, key, directory, decompress, entry)| async move {
-                let (path, local_etag, metadata) = entry;
+                let MappedPathResult {
+                    source_path,
+                    local_etag,
+                    metadata,
+                    ..
+                } = entry;
                 let metadata = metadata.unwrap();
                 match local_etag {
                     Ok(local_etag) => {
                         if local_etag != metadata.e_tag {
                             debug!(
                                 "etag mismatch: {}, local etag={}, remote etag={}",
-                                path.display(),
+                                source_path.display(),
                                 local_etag,
                                 metadata.e_tag
                             );
@@ -562,12 +595,12 @@ where
                             )
                             .await?;
                         } else {
-                            debug!("etag match: {}", path.display());
+                            debug!("etag match: {}", source_path.display());
                         }
                     }
                     Err(err) => match err {
                         Error::ETagNotPresent => {
-                            debug!("file did not exist locally: {}", path.display());
+                            debug!("file did not exist locally: {}", source_path.display());
                             download_with_dir(
                                 &s3,
                                 &bucket,
@@ -593,9 +626,8 @@ async fn sync_remote_to_local<T>(
     bucket: &str,
     key: &str,
     directory: impl AsRef<Path>,
-    glob_includes: &[Pattern],
-    glob_excludes: &[Pattern],
-    decompress: bool,
+    filters: &[GlobFilter],
+    transparent_decompress: bool,
 ) -> Result<()>
 where
     T: S3 + Sync + Send + Clone,
@@ -603,22 +635,22 @@ where
     let directory = directory.as_ref();
     let task_count = Config::global().concurrent_sync_tasks();
     let object_listing =
-        flattened_object_listing(s3, bucket, key, directory, glob_includes, glob_excludes);
+        flattened_object_listing(s3, bucket, key, directory, filters, transparent_decompress);
 
-    let compression = if decompress {
+    let compression = if transparent_decompress {
         SyncCompressionDirection::Down
     } else {
         SyncCompressionDirection::None
     };
 
-    let etag_stream = map_paths_to_etags(object_listing, compression).buffer_unordered(task_count);
+    let etag_stream = translate_paths(object_listing, compression).buffer_unordered(task_count);
     let sync_tasks = remote_to_local_sync_tasks(
         s3.clone(),
         bucket.into(),
         key.into(),
         directory.into(),
         etag_stream,
-        decompress,
+        transparent_decompress,
     );
 
     sync_tasks
@@ -629,26 +661,81 @@ where
     Ok(())
 }
 
+async fn download_with_dir<T>(
+    s3: &T,
+    bucket: &str,
+    s3_prefix: &str,
+    s3_suffix: &str,
+    local_dir: impl AsRef<Path>,
+    transparent_decompress: bool,
+) -> Result<()>
+where
+    T: S3 + Sync + Send + Clone,
+{
+    let local_dir = local_dir.as_ref();
+    let dest_path = local_dir.join(s3_suffix);
+
+    let parent_dir = dest_path.parent().ok_or(Error::ParentDirNone)?;
+    fs::create_dir_all(parent_dir).await?;
+
+    let key = format!("{}", Path::new(s3_prefix).join(s3_suffix).display());
+    let dest_path = format!("{}", dest_path.display());
+
+    if transparent_decompress {
+        crate::download_with_transparent_decompression(s3, bucket, &key, &dest_path).await?;
+    } else {
+        crate::download(s3, bucket, &key, &dest_path).await?;
+    }
+
+    Ok(())
+}
+
 #[cfg(test)]
 mod tests {
     use super::*;
 
     #[test]
     fn test_process_globs() {
-        let includes = vec![Pattern::new("*.csv").unwrap()];
-        let excludes = vec![Pattern::new("*-blah.csv").unwrap()];
+        let filters = vec![
+            GlobFilter::Exclude(Pattern::new("*-blah.csv").unwrap()),
+            GlobFilter::Include(Pattern::new("*.csv").unwrap()),
+        ];
 
-        assert!(process_globs("data.sbp", &includes[..], &excludes[..]).is_none());
-        assert!(process_globs("yes.csv", &includes[..], &excludes[..]).is_some());
-        assert!(process_globs("no-blah.csv", &includes[..], &excludes[..]).is_none());
+        assert!(process_globs("data.sbp", &filters[..]).is_none());
+        assert!(process_globs("yes.csv", &filters[..]).is_some());
+        assert!(process_globs("no-blah.csv", &filters[..]).is_none());
     }
 
     #[test]
     fn test_process_globs_exclude_all() {
-        let includes = vec![Pattern::new("*.png").unwrap()];
-        let excludes = vec![];
+        let filters = vec![GlobFilter::Include(Pattern::new("*.png").unwrap())];
 
-        assert!(process_globs("a-fancy-thing.png", &includes[..], &excludes[..]).is_some());
-        assert!(process_globs("horse.gif", &includes[..], &excludes[..]).is_none());
+        assert!(process_globs("a-fancy-thing.png", &filters[..]).is_some());
+        assert!(process_globs("horse.gif", &filters[..]).is_none());
+    }
+
+    #[test]
+    fn test_process_globs_explicit_exclude() {
+        let filters = vec![
+            GlobFilter::Include(Pattern::new("*.png").unwrap()),
+            GlobFilter::Exclude(Pattern::new("*.txt").unwrap()),
+            GlobFilter::Include(Pattern::new("*").unwrap()),
+        ];
+
+        assert!(process_globs("a-fancy-thing.png", &filters[..]).is_some());
+        assert!(process_globs("horse.gif", &filters[..]).is_some());
+        assert!(process_globs("myfile.txt", &filters[..]).is_none());
+    }
+
+    #[test]
+    fn test_process_globs_precedence() {
+        let filters = vec![
+            GlobFilter::Exclude(Pattern::new("*").unwrap()),
+            GlobFilter::Include(Pattern::new("*").unwrap()),
+        ];
+
+        assert!(process_globs("a-fancy-thing.png", &filters[..]).is_none());
+        assert!(process_globs("horse.gif", &filters[..]).is_none());
+        assert!(process_globs("myfile.txt", &filters[..]).is_none());
     }
 }
